RECORD_DFT_ATTR = {
    'name',
    'created',
    'msecs',
    'relativeCreated',
    'asctime',
    'levelname',
    'levelno',
    'pathname',
    'filename',
    'module',
    'lineno',
    'funcName',
    'thread',
    'threadName',
    'processName',
    'process',
    'exc_info',
    'exc_text',
    'stack_info',
    'msg',
    'args',
    'message'
}
